#pragma once

template <class T, int SIZE>
class Median {
public:
	Median() : _Index(0), _Count(0) {
		_Buffer[0] = 0;
		_Buffer[SIZE - 1] = 0;
	}

	T GetMedian() {
		register int i, j;

		for (size_t i = 0; i < SIZE; i++) {
			_Buffer[SIZE + i] = _Buffer[i];
		}

		int k = _Count & 1 ? _Count / 2 : _Count / 2 - 1;
		k += SIZE;

		register int l = SIZE;
		int m = SIZE + _Count - 1;

		while (l < m) {
			T x = _Buffer[k];
			i = l;
			j = m;
			do {
				while (_Buffer[i] < x) {
					i++;
				}
				while (x < _Buffer[j]) {
					j--;
				}
				if (i <= j) {
					Swap(i, j);
					i++;
					j--;
				}
			} while (i <= j);
			if (j < k) l = i;
			if (k < i) m = j;
		}
		return _Buffer[k];
	}

	const T GetQuantile(double p) {
		p = p < 0 ? 0 : (p > 1 ? 1 : p);
		for (size_t i = 0; i < SIZE; i++) {
			_Buffer[SIZE + i] = _Buffer[i];
		}
		Sort();
		unsigned int i = (_Count * p) - 1; // n * p
		// Vereinfachte Form des Quantils (geht von nicht ganzzahligem n * p aus)
		return _Buffer[SIZE + i];
	}
	
	T GetAverage(unsigned int Count = 0xffffffff) {
		GetMedian();
		
		if (Count > _Count) {
			Count = _Count;
		}

		double avg = 0;

		for (size_t i = _Count / 2 - Count / 2; i <= _Count / 2 + Count / 2; i++) {
			avg += _Buffer[SIZE + i];
		}
		avg /= Count;

		return (T) avg;
	}

	T& operator +=(const T Data) {
		return Add(Data);
	}

	void Clear() {
		_Index = 0;
		_Count = 0;
	}

private:
	T _Buffer[SIZE * 2];
	size_t _Index;
	size_t _Count;

	T& Add(const T Data) {
		_Buffer[_Index] = Data;
		T& Result = _Buffer[_Index];

		if (++_Index >= SIZE) {
			_Index = 0;
		}
		if (++_Count >= SIZE) {
			_Count = SIZE;
		}
		return Result;
	}

	int GetPrevious() {
		return _Index > 0 ? _Index - 1 : SIZE - 1;
	}

	inline void Swap(unsigned int i, unsigned int j) {
		const T t = _Buffer[i];
		_Buffer[i] = _Buffer[j];
		_Buffer[j] = t;
	}


	// Shell Sort nach Marcin Ciura

	void Sort() {
		const unsigned int Gaps[] = {
			701, 301, 132, 57, 23, 10, 4, 1
		};

		for (unsigned int Gap = 0; Gap < sizeof(Gaps) / sizeof(unsigned int); Gap++) {
			for (unsigned int i = Gap; i < _Count; i++) {
				T Temp = _Buffer[i];
				unsigned int j = i;
				for (j = i; j >= Gap && _Buffer[j - Gap] > Temp; j -= Gap) {
					_Buffer[j] = _Buffer[j - Gap];
				}
				_Buffer[j] = Temp;
			}
		}
	}
};