#define LTC2990_DEBUG

#include <i2c_t3.h>
#include "Median.h"
//#include <Wire.h>
//#include <LTC2990.h>

#define LTC2990_MEASURE_ONLY_INT 0x0
#define LTC2990_MEASURE_ONLY_V1_TR1 0x1
#define LTC2990_MEASURE_ONLY_V3_TR2 0x2
#define LTC2990_MEASURE_ALL 0x3

#define LTC2990_MEASURE_MODE_V1_V2_TR2 0x0
#define LTC2990_MEASURE_MODE_V1_MINUS_V2_TR2 0x1
#define LTC2990_MEASURE_MODE_V1_MINUS_V2_V3_V4 0x2
#define LTC2990_MEASURE_MODE_TR1_V3_V4 0x3
#define LTC2990_MEASURE_MODE_TR1_V3_MINUS_V4 0x4
#define LTC2990_MEASURE_MODE_TR1_TR2 0x5
#define LTC2990_MEASURE_MODE_V1_MINUS_V2_V3_MINUS_V4 0x6
#define LTC2990_MEASURE_MODE_V1_V2_V3_V4 0x7

#define LTC2990_CONTROL_TEMP_CELSIUS 0x00
#define LTC2990_CONTROL_TEMP_KELVIN 0x80
#define LTC2990_CONTROL_TRIGGER_REPEATED 0x00
#define LTC2990_CONTROL_TRIGGER_SINGLE 0x40

#define LTC2990_REGISTER_STATUS 0x0
#define LTC2990_REGISTER_CONTROL 0x1
#define LTC2990_REGISTER_TRIGGER 0x2
#define LTC2990_REGISTER_NA 0x3
#define LTC2990_REGISTER_TINT_M 0x4
#define LTC2990_REGISTER_TINT_L 0x5
#define LTC2990_REGISTER_V1_M 0x6
#define LTC2990_REGISTER_V1_L 0x7
#define LTC2990_REGISTER_V2_M 0x8
#define LTC2990_REGISTER_V2_L 0x9
#define LTC2990_REGISTER_V3_M 0xA
#define LTC2990_REGISTER_V3_L 0xB
#define LTC2990_REGISTER_V4_M 0xC
#define LTC2990_REGISTER_V4_L 0xD
#define LTC2990_REGISTER_VCC_M 0xE
#define LTC2990_REGISTER_VCC_L 0xF

#define LTC2990_ADDRESS 0x4c

#define LTC2990_BIT_VALID 0x80
#define LTC2990_BIT_SIGN 0x40

#define LTC2990_VOLTAGE_SINGLE_END_COEF 305.180 * 5.55486759548144
#define LTC2990_VOLTAGE_DIFFERENTIAL_COEF 19.420
#define LTC2990_TEMP_COEF 0.0625

#define LTC2990_STATUS_BUSY_BIT 0x0
#define LTC2990_STATUS_READY_TINT_BIT 0x1
#define LTC2990_STATUS_READY_V1_BIT 0x2
#define LTC2990_STATUS_READY_V2_BIT 0x3
#define LTC2990_STATUS_READY_V3_BIT 0x4
#define LTC2990_STATUS_READY_V4_BIT 0x5
#define LTC2990_STATUS_READY_VCC_BIT 0x6
#define LTC2990_STATUS_RESERVED_BIT 0x7

#define SINGLE_COUNTS_TO_VOLTAGE(CH, C) (CH == 0 ? (0.001694915254 * C + 0.006779661017) : (0.001167133520 * C + 0.002334267))

int16_t VoltageRaw[4] = { 0 };
int16_t VoltageDiffRaw[4] = { 0 };

bool WriteToLtc2990(uint8_t Address, uint8_t Register, uint8_t Value) {
	noInterrupts();
	Wire.beginTransmission(Address);
	Wire.write(Register);
	Wire.write(Value);
	uint8_t Status = Wire.endTransmission(I2C_STOP);
	interrupts();

	if (Status != 0) {
		Serial.printf("Write Error %d\n", Status);
		return false;
	}
	return true;
}

bool ReadFromLtc2990(uint8_t Address, uint8_t Register, uint8_t& Value) {
	noInterrupts();
	Wire.beginTransmission(Address);
	Wire.write(Register);
	uint8_t Status = Wire.endTransmission(true);

	if (Status == 0) {
		if (Wire.requestFrom(Address, 1, true) == 1) {
			Value = Wire.read();
			interrupts();
			return true;
		}
	}
	interrupts();
	return false;
}

bool SetLtc2990Mode(bool IsCelsiusTemperature, bool SingleTrigger, uint8_t Limit, uint8_t Mode) {
	uint8_t Control = 0;

	if (IsCelsiusTemperature == false) {
		Control |= LTC2990_CONTROL_TEMP_KELVIN;
	}

	if (SingleTrigger) {
		Control |= LTC2990_CONTROL_TRIGGER_SINGLE;
	}

	Control |= Limit << 3;
	Control |= Mode;

	return WriteToLtc2990(LTC2990_ADDRESS, LTC2990_REGISTER_CONTROL, Control);
}

bool ReadSingleVoltageFromLtc2990(uint8_t Address, uint8_t Channel, float& Voltage) {
	uint8_t r[2] = { 0 };

	if (ReadFromLtc2990(Address, LTC2990_REGISTER_V1_M + Channel * 2, r[0]) == false || ReadFromLtc2990(Address, LTC2990_REGISTER_V1_L + Channel * 2, r[1]) == false) {
		return false;
	}

	if (r[0] & LTC2990_BIT_VALID) {
		int16_t raw = (r[0] & 0x3f) << 8 | r[1];
		if (r[0] & LTC2990_BIT_SIGN) {
			raw = -((16384 - raw) + 1);
		}
		VoltageRaw[Channel] = raw;
		Voltage = (float) raw * LTC2990_VOLTAGE_SINGLE_END_COEF / 1000000.0;
		return true;
	}
	return false;
}

bool ReadDifferentialVoltageFromLtc2990(uint8_t Address, uint8_t Channel, float& Voltage) {
	uint8_t r[2] = { 0 };

	if (ReadFromLtc2990(Address, LTC2990_REGISTER_V1_M + Channel * 2, r[0]) == false || ReadFromLtc2990(Address, LTC2990_REGISTER_V1_L + Channel * 2, r[1]) == false) {
		return false;
	}

	if (r[0] & LTC2990_BIT_VALID) {
		int16_t raw = ((r[0] & 0x3f) << 8 | r[1]);
		if (r[0] & LTC2990_BIT_SIGN) {
			raw = -((16384 - raw) + 1);
		}
		VoltageDiffRaw[Channel] = raw;
		Voltage = (float) raw * LTC2990_VOLTAGE_DIFFERENTIAL_COEF / 1000000.0;
		return true;
	}
	return false;
}

bool ReadVccFromLtc2990(uint8_t Address, float& Vcc) {
	uint8_t r[2] = { 0 };

	if (ReadFromLtc2990(LTC2990_ADDRESS, LTC2990_REGISTER_VCC_M, r[0]) == false || ReadFromLtc2990(LTC2990_ADDRESS, LTC2990_REGISTER_VCC_L, r[1]) == false) {
		return false;
	}

	if (r[0] & LTC2990_BIT_VALID) {
		int16_t raw = r[1] | ((r[0] & 0x3f) << 8);
		if (r[0] & LTC2990_BIT_SIGN) {
			raw = -((16384 - raw) + 1);
		}
		Vcc = 2.5 + (float) raw * 0.00030518;
		return true;
	}
	return false;
}

bool ReadInternalTemperatureFromLtc2990(uint8_t Address, float& Temp) {
	uint8_t r[2] = { 0 };

	if (ReadFromLtc2990(LTC2990_ADDRESS, LTC2990_REGISTER_TINT_M, r[0]) == false || ReadFromLtc2990(LTC2990_ADDRESS, LTC2990_REGISTER_TINT_L, r[1]) == false) {
		return false;
	}

	uint16_t raw = r[1] | ((r[0] & 0x1f) << 8);

	Temp = (float) raw * LTC2990_TEMP_COEF;
	return true;
}

bool WaitForLtc2990Status(uint8_t Address, uint8_t Mask, uint32_t Timeout) {
	uint8_t Status = 0;

	digitalWrite(LED_BUILTIN, HIGH);
	
	WriteToLtc2990(Address, LTC2990_REGISTER_TRIGGER, 1);

	int t = millis();
	do {
		ReadFromLtc2990(Address, LTC2990_REGISTER_STATUS, Status);
		
		if (millis() - t > Timeout) {
			return false;
		}
	} while ((Status & Mask) != Mask);

	digitalWrite(LED_BUILTIN, LOW);

	return true;
}

void setup() {
	delay(1);

	Serial.begin(9600);

	pinMode(LED_BUILTIN, OUTPUT);

	Wire.begin();
	//  Wire.resetBus();
	//  Wire.setRate(I2C_RATE_400);
	Wire.pinConfigure(I2C_PINS_18_19, I2C_PULLUP_EXT);
	Wire.setTimeout(20000);
	Wire.setOpMode(I2C_OP_MODE_IMM);

}

Median<int16_t, 9> _RawSingleEnded[4];
Median<double, 11> _SingleEnded[4];

void loop() {
	uint8_t Mask = (1 << LTC2990_STATUS_READY_V1_BIT) | (1 << LTC2990_STATUS_READY_V2_BIT) | (1 << LTC2990_STATUS_READY_V3_BIT) | (1 << LTC2990_STATUS_READY_V4_BIT); // | (1 << LTC2990_STATUS_READY_VCC_BIT) | (1 << LTC2990_STATUS_READY_TINT_BIT);
	uint8_t Status = 0;

	SetLtc2990Mode(true, false, LTC2990_MEASURE_ALL, LTC2990_MEASURE_MODE_V1_V2_V3_V4);

	if (WaitForLtc2990Status(LTC2990_ADDRESS, Mask, 150) == false) {
		Serial.printf("WaitForStatus(SINGLE) failed.\n");
	}

	float SingleEnded[4] = { 0 };

	ReadSingleVoltageFromLtc2990(LTC2990_ADDRESS, 0, SingleEnded[0]); // V1
	ReadSingleVoltageFromLtc2990(LTC2990_ADDRESS, 1, SingleEnded[1]); // V2
	ReadSingleVoltageFromLtc2990(LTC2990_ADDRESS, 2, SingleEnded[2]); // V3
	ReadSingleVoltageFromLtc2990(LTC2990_ADDRESS, 3, SingleEnded[3]); // V4

	SetLtc2990Mode(true, false, LTC2990_MEASURE_ALL, LTC2990_MEASURE_MODE_V1_MINUS_V2_V3_MINUS_V4);

	if (WaitForLtc2990Status(LTC2990_ADDRESS, Mask, 150) == false) {
		Serial.printf("WaitForStatus(DIFF) failed.\n");
	}
	
	float Diff[2] = { 0 };

	ReadDifferentialVoltageFromLtc2990(LTC2990_ADDRESS, 0, Diff[0]); // V1-V2
	ReadDifferentialVoltageFromLtc2990(LTC2990_ADDRESS, 2, Diff[1]); // V3-V4

	_RawSingleEnded[0] += VoltageRaw[0];
	_RawSingleEnded[1] += VoltageRaw[1];
	_RawSingleEnded[2] += VoltageRaw[2];
	_RawSingleEnded[3] += VoltageRaw[3];

	_SingleEnded[0] += SingleEnded[0];
	_SingleEnded[1] += SingleEnded[1];
	_SingleEnded[2] += SingleEnded[2];
	_SingleEnded[3] += SingleEnded[3];

	//Serial.printf("V1: (%8d) %8.4f, V2: (%8d) %8.4f, V3: (%8d) %8.4f, V4: (%8d) %8.4f\n",
	//	_RawSingleEnded[0].GetAverage(),
	//	SINGLE_COUNTS_TO_VOLTAGE(0, _RawSingleEnded[0].GetAverage()),
	//	_RawSingleEnded[1].GetAverage(),
	//	SINGLE_COUNTS_TO_VOLTAGE(0, _RawSingleEnded[1].GetAverage()),
	//	_RawSingleEnded[2].GetAverage(),
	//	SINGLE_COUNTS_TO_VOLTAGE(1, _RawSingleEnded[2].GetAverage()),
	//	_RawSingleEnded[3].GetAverage(),
	//	SINGLE_COUNTS_TO_VOLTAGE(1, _RawSingleEnded[3].GetAverage())
	//);
	Serial.printf("V1 %8.4f V2 %8.4f V3 %8.4f V4 %8.4f\n", _SingleEnded[0].GetAverage(), _SingleEnded[1].GetAverage(), SingleEnded[2], SingleEnded[3]);
	//Serial.printf("%d	%d	%d	%d	%d	%d\n", VoltageRaw[0], VoltageRaw[1], VoltageRaw[2], VoltageRaw[3], VoltageDiffRaw[0], VoltageDiffRaw[2]);
}

